//Copyright 2023 (C) Brian William Denton
//GNU GPL-3 LICENSE

#define _GNU_SOURCE

#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/stat.h>
#include <time.h>
#include <errno.h>

#include <stdlib.h>
#include <stdio.h>


union arrow
{
	unsigned char str[8];
	uint64_t uint;
};

#ifdef __ORDER_LITTLE_ENDIAN__
const union arrow up = {.uint = 0x9880e2a080e2a2c3};
const union arrow down = {.uint = 0x9c80e2a080e2a2c3};
#else
const union arrow up = {.uint = 0xc3a2e280a0e28098};
const union arrow down = {.uint = 0xc3a2e280a0e2809c};
#endif

char *msgclass[] = {"down", "up"};

struct smsmsg
{
	time_t time;
	unsigned char isup;
	struct smsmsg *l;
	struct smsmsg *r;
	struct tm tmtime;
	char msg[];
};

#define DENT_BUF_SIZE 4096
char dents[DENT_BUF_SIZE];
int indirfd;
int ofd;

int myerrno;

#define FINRAM_DEFAULT_SIZE 128
char *finram;
size_t finramsize;

struct smsmsg *messages = NULL;

char *phonenum;

static inline void printusage()
{
	puts("sms-extractor DIRECTORY PHONE_NUMBER_STRING OUTPUT_FILE.htmlfrag\n\nCopyright (C) 2023 Brian William Denton\nGNU GPL-3 License");
	return;
}

static inline void addmsg(struct smsmsg *m)
{
	struct smsmsg *cur;

	m->l = NULL;
	m->r = NULL;

	if (!messages)
	{
		messages = m;
		return;
	}

	cur = messages;
	while (1)
	{
		if (m->time < cur->time)
		{
			if (cur->l)
			{
				cur = cur->l;
				continue;
			}
			cur->l = m;
			return;
		}
		if (cur->r)
		{
			cur = cur->r;
			continue;
		}
		cur->r = m;
		return;
	}
}

static inline unsigned countnls(char *s)
{
	char *curpos;
	char *newpos;
	unsigned retval = 0;

	curpos = s;
	while (1)
	{
		newpos = __builtin_strchr(curpos, '\n');
		if (newpos)
		{
			retval++;
			curpos = newpos+1;
		}
		else
		{
			break;
		}
	}

	return retval;
}

static inline void processmsg(char *fname)
{
	int fd;
	struct smsmsg *newmsg;
	char *msgcontents;
	size_t msglen;
	char *endptr;
	char *nlpos;
	struct stat fst;
	unsigned nls;
	unsigned inspoint;

	fd = openat(indirfd, fname, O_RDONLY);
	if (fd == -1)
	{
		myerrno = errno;
		perror("src open()");
		exit(myerrno);
	}
	if (fstat(fd, &fst) == -1)
	{
		myerrno = errno;
		perror("fstat()");
		exit(myerrno);
	}

	if (!fst.st_size)
	{
		puts("empty file!");
		exit(1);
	}

	//or equal too so I have room for null!
	//NEVERMIND I NOW NULL THE TERMINATING NEWLINE!
	if ((size_t)fst.st_size > finramsize)
	{
		free(finram);
		finramsize = fst.st_size;
		finram = malloc(finramsize);
		if (!finram)
		{
			myerrno = errno;
			perror("malloc()");
			exit(myerrno);
		}
	}

	if (read(fd, finram, fst.st_size) != fst.st_size)
	{
		myerrno = errno;
		perror("read()");
		exit(myerrno);
	}
	finram[fst.st_size-1] = 0;
	close(fd);

	msgcontents = __builtin_strchr(finram, '\n');
	if (!(*msgcontents))
	{
		puts("bad file!");
		exit(1);
	}
	msgcontents += 2;

	nls = countnls(msgcontents);
	
	msglen = ((fst.st_size - (msgcontents - finram))  + (4 * nls));
	newmsg = malloc(sizeof(struct smsmsg) + msglen);
	if (!newmsg)
	{
		myerrno = errno;
		perror("malloc()");
		exit(myerrno);
	}

	if (!nls)
	{
		__builtin_memcpy(newmsg->msg, msgcontents, msglen);
	}
	else
	{
		inspoint = 0;
		endptr = msgcontents;
		while (1)
		{
			nlpos = __builtin_strchr(endptr, '\n');
			if (!nlpos)
			{
				break;
			}
			__builtin_memcpy(&newmsg->msg[inspoint], endptr, (nlpos - endptr));
			inspoint += (nlpos - endptr);
			endptr = (nlpos + 1);
			__builtin_memcpy(&newmsg->msg[inspoint], "<br>", 4);
			inspoint += 4;
		}
		__builtin_memcpy(&newmsg->msg[inspoint], endptr, msglen - inspoint);
	}

	newmsg->isup = ((*((uint64_t*)(&fname[20]))) == up.uint);

	fname[19] = 0;
	__builtin_memset(&newmsg->tmtime, 0, sizeof(struct tm));
	endptr = strptime(fname, "%Y-%m-%d %H.%M.%S", &newmsg->tmtime);
	if ((!endptr) || ((*endptr) != 0))
	{
		puts("strptime() failed");
		exit(1);
	}
	newmsg->time = mktime(&newmsg->tmtime);
	if (newmsg->time == -1)
	{
		myerrno = errno;
		perror("mktime()");
		exit(myerrno);
	}

	addmsg(newmsg);
}

static inline void dodir()
{
	ssize_t dentretval;
	unsigned curdentpos;
	struct dirent64 *curdent;
	char *numend;
	char orignumend;

	while (1)
	{
		dentretval = getdents64(indirfd, dents, DENT_BUF_SIZE);
		if (dentretval == -1)
		{
			myerrno = errno;
			perror("getdents64()");
			exit(myerrno);
		}
		if (dentretval == 0)
		{
			break;
		}
		curdentpos = 0;
		while (curdentpos != dentretval)
		{
			curdent = (struct dirent64*)&dents[curdentpos];
			if ((curdent->d_type == DT_REG) || (curdent->d_type == DT_LNK))
			{
				if (__builtin_strlen(curdent->d_name) < 34)
				{
					puts(curdent->d_name);
					puts("filename too short");
					exit(1);
				}
				else
				{
					//puts(&curdent->d_name[29]);
					numend = __builtin_strchr(&curdent->d_name[29], ' ');
					if (!numend)
					{
						numend = (curdent->d_name) + (__builtin_strlen(curdent->d_name) - 4);
						orignumend = '.';
					}
					else
					{
						orignumend = ' ';
					}
					*numend = 0;
					//puts(&curdent->d_name[29]);
					if (!__builtin_strcmp(&curdent->d_name[29], phonenum))
					{
						//found number!
						*numend = orignumend;
						processmsg(curdent->d_name);
					}
				}
			}
			curdentpos += curdent->d_reclen;
		}
	}
	return;
}

/* void printrecursor(struct smsmsg *m) */
/* { */
/* 	if (m->l) */
/* 	{ */
/* 		printrecursor(m->l); */
/* 	} */
/* 	printf("%lu %hhu - \"%s\"\n", m->time, m->isup, m->msg); */
/* 	if (m->r) */
/* 	{ */
/* 		printrecursor(m->r); */
/* 	} */
/* } */

void writerecursor(struct smsmsg *m)
{
	if (m->l)
	{
		writerecursor(m->l);
	}
	if (dprintf(ofd, "\t\t\t\t<div class=\"%s\"><p>%s</p><p class=\"date\">%i-%02i-%02i %02i:%02i:%02i</p></div>\n", msgclass[m->isup], m->msg, 1900+m->tmtime.tm_year, 1+m->tmtime.tm_mon, m->tmtime.tm_mday, m->tmtime.tm_hour, m->tmtime.tm_min, m->tmtime.tm_sec) < 0)
	{
		myerrno = errno;
		perror("dprintf()");
		exit(myerrno);
	}
	if (m->r)
	{
		writerecursor(m->r);
	}
}

static inline void dooutput()
{
	if (dprintf(ofd, "\t\t\t<div class=\"mthread\">\n\t\t\t\t<div class=\"contact\"><h1>%s</h1></div>\n", phonenum) < 0)
	{
		myerrno = errno;
		perror("dprintf()");
		exit(myerrno);
	}
	writerecursor(messages);
	if (write(ofd, "\t\t\t</div>\n", 10) != 10)
	{
		myerrno = errno;
		perror("write()");
		exit(myerrno);
	}
}

int main(int argc, char *argv[])
{
	if (argc != 4)
	{
		printusage();
		return EINVAL;
	}

	finramsize = FINRAM_DEFAULT_SIZE;
	finram = malloc(finramsize);
	if (!finram)
	{
		myerrno = errno;
		perror("malloc()");
		return myerrno;
	}

	indirfd = open(argv[1], O_RDONLY | O_DIRECTORY);
	if (indirfd == -1)
	{
		myerrno = errno;
		perror("directory open()");
		return myerrno;
	}

	phonenum = argv[2];

	ofd = open(argv[3], O_WRONLY | O_EXCL | O_CREAT, 0644);
	if (ofd == -1)
	{
		myerrno = errno;
		perror("output_file.txt open()");
		close(indirfd);
		return myerrno;
	}

	dodir();
	close(indirfd);

	if (!messages)
	{
		puts("did not find any messages!");
		close(ofd);
		if (unlink(argv[3]) == -1)
		{
			myerrno = errno;
			perror("output file unlink()");
			return myerrno;
		}
	}
	else
	{
		dooutput();
		close(ofd);
	}

	return 0;
}
